# -*- coding: utf-8 -*-
'''Object wrapper for tct'''
import os
import sys
import marshal

import tokyo.cabinet as tc

from fields import *


class class_cached_property(object):
    '''
    Кеширующий classmethod
    '''
    def __init__(self, f):
        self.f = f

    def __get__(self, instance, owner):
        value = self.f(owner)
        setattr(owner, self.f.__name__, value)
        return value


class Error(Exception):
    pass


class Model(object):
    '''
    Если у документа нет 

    '''
    __filename__ = None
    key = None
    
    def __init__(self, key=None, _raw_dict=None, **kwargs):
        if key is not None:
            self.key = key if isinstance(key, unicode) else key.decode('utf-8')
        if _raw_dict:
            self._data = _raw_dict
        else:
            self._data = {}
            for prop_name, value in kwargs.iteritems():
                if prop_name not in self._fields:
                    assert 0, 'Unknown field "%s" in model "%s"' % (
                        prop_name, self.__class__.__name__)
                setattr(self, prop_name, value)
    
    def save(self):
        if self.key is None:
            self.key = unicode(len(self.db) + 1)
        #~ print 'self.db[%s] = %s' % (self.key, self._data)
        self.db[self.key] = self._data
        
    @classmethod
    def get(cls, key):
        key = key if isinstance(key, unicode) else key.decode('utf-8')
        try:
            return cls(key, cls.db[key])
        except KeyError:
            return None
            
    @classmethod
    def filter(cls, operator=None, value=None):
        ret = Filter(cls)
        if operator:
            ret = ret.filter(operator, value)
        return ret

    def delete(self):
        if not self.key:
            assert 0, 'the object is not saved'
        del self.db[self.key]
        self._data = {}

    @class_cached_property
    def db(cls):
        '''Отложенное подключение к бд'''
        if not cls.__filename__:
            assert 0, 'set __filename__ in model "%s"' % cls.__name__
        db = tc.TDB()
        db.open(cls.__filename__, tc.HDBOWRITER | tc.HDBOCREAT)
        for prop in (cls.__dict__[n] for n in cls._fields):
            #~ print prop.__class__.__name__
            if prop.index:
                #~ print prop.__class__.__name__, prop.INDEX
                if not prop.INDEX:
                    raise Error('field "%s" is not indexed' % prop.__class__.__name__)
                #~ print 'cls.db.setindex(%s, %s)' % (prop.key, prop.INDEX['set'])
                db.setindex(prop.key, prop.INDEX['set'])
            else:
                try:
                    db.setindex(prop.key, tc.TDBITVOID)
                except tc.Error:
                    pass
        return db
    
    @class_cached_property
    def _fields(cls):
        '''
        {prop_attr_name: save_key_name}
        '''
        def rec_bases(cls):
            bases = list(cls.__bases__) or []
            ret = []
            for base in bases:
                for c in rec_bases(base):
                    if c not in ret and c not in bases:
                        ret.append(c)
            ret.extend(bases)
            return ret
        ret = {}
        for k, prop in cls.__dict__.items():
            if BaseField in rec_bases(prop.__class__):
                if not prop.key:
                    prop.key = k
                if prop.key in ret.values():
                    assert 0, 'multiply key "%s" in model "%s"' % (prop.key, cls.__name__)
                ret[k] = prop.key
        return ret
        
    def __unicode__(self):
        items = []
        for k in ['key'] + self._fields.keys():
            v = getattr(self, k)
            if isinstance(v, str):
                v = "'%s'" % v
            if isinstance(v, unicode):
                v = "u'%s'" % v
            try:
                v = unicode(v)
                if len(v) > 50:
                    v = v[:50] + ' ...'
            except UnicodeDecodeError:
                v = '- not printable -'
            items.append((k, v))
        max_len = max([len(k) for k in dict(items).keys()])
        ret = self.__class__.__name__ + '(\n'
        tpl = '    %%-%is = %%s' % max_len
        return ret + ',\n'.join([tpl % (k, v) for k, v in items]) + '\n)'
            
    def __str__(self):
        return self.__unicode__().encode('utf-8')


class Filter(object):
    def __init__(self, model, queue=None, order=None, limit=None):
        self.model = model
        self.queue = queue or []
        self._order = order
        self._limit = limit
    
    def filter(self, operator, value):
        field, condition_abbr = operator.split()
        try:
            prop = self.model.__dict__[field]
        except KeyError:
            assert 0, 'unknown field: "%s"' % field
        try:
            condition = prop.get_condition(condition_abbr)
        except KeyError:
            assert 0, 'unknown condition: "%s" for "%s" field' % (
                condition_abbr, field)
        queue = self.queue[:]
        queue.append((prop.key, condition, prop.dumps(value)))
        return Filter(self.model, queue, self._order, self._limit)
    
    def order(self, field):
        direction = 'asc'
        if field.startswith('-'):
            field = field[1:]
            direction = 'desc'
        try:
            prop = self.model.__dict__[field]
        except KeyError:
            assert 0, 'unknown field: "%s"' % field
        try:
            #~ print prop.__class__.__name__, prop.key
            order = (prop.key, prop.get_order(direction))
        except KeyError:
            assert 0, 'not ordered field: "%s"' % field
        return Filter(self.model, self.queue, order, self._limit)
        
    def __getitem__(self, key):
        '''Limit'''
        start = 0
        if isinstance(key, slice):
            start = key.start or 0
            if key.stop:
                limit = (key.stop - start, start)
            elif start:
                limit = (-1, start)
        else:
            limit = (1, key)
        return Filter(self.model, self.queue, self._order, limit)

    def _query(self):
        q = self.model.db.query()
        for args in self.queue:
            q.filter(*args)
        if self._order:
            #~ print 'order:', self._order
            q.sort(*self._order)
        if self._limit:
            q.limit(*self._limit)
        return q

    def __iter__(self):
        for k in self._query().search():
            yield self.model(k, _raw_dict=self.model.db[k])
    
    def count(self):
        return len(self._query().search())

    def delete(self):
        self._query().remove()
        
    def get(self):
        for k in self[0]:
            return k
        return None

    def __str__(self):
        return '\nfilters: %s\norder: %s\nlimit: %s' % (
            self.queue, self._order, self._limit)


class A(Model):
    number      = IntegerField(key='n', index=True)
    fnum        = FloatField()
    name        = UnicodeField(index=True)
    birthday    = DateField(key='b', index=True)
    created     = DateTimeField(index=True)
    #~ tst         = ByteStringField(index=True)
    
    __filename__ = '/home/imbolc/0/var/otct.test.tct'


#~ A.db.clear()

#~ for i in xrange(100):
    #~ A(
        #~ number  = i,
        #~ name    = u'ЫЫЫ',
        #~ fnum    = -1.0 / (i + 1),
        #~ birthday= date(1900 + i, 05, 06),
        #~ created = datetime.now()
    #~ ).save()

#~ print '--- 2'
#~ a.save()
#~ print a

#~ A.db.optimize()
#~ print A.db.uid()

#~ f = A.filter('number >=', 10)
#~ f = f.filter('birthday <', date(1930, 1, 1))
#~ print f.count()
#~ for obj in f.order('fnum'):
#~ for obj in f.order('-number'):
    #~ print obj

#~ f.delete()

#~ a = A.get(6)
#~ print a
#~ a.delete()

if __name__ == "__main__":
    import doctest
    doctest.testmod()
    



# -*- coding: utf-8 -*-
import sys
import struct
import time
from datetime import datetime, date

import tokyo.cabinet as tc


INDEX_STR = {'set': tc.TDBITLEXICAL, 'asc': tc.TDBQOSTRASC,
            'desc': tc.TDBQOSTRDESC}
INDEX_NUM = {'set': tc.TDBITDECIMAL, 'asc': tc.TDBQONUMASC,
            'desc': tc.TDBQONUMDESC}    
CONDITIONS = {
    'str': {
        '=' : tc.TDBQCSTREQ,
        'in' : tc.TDBQCSTRINC,
        'startswith' : tc.TDBQCSTRBW,
        'endswith' : tc.TDBQCSTREW,
    },
    'num': {
        '=' : tc.TDBQCNUMEQ,
        '>' : tc.TDBQCNUMGT,
        '>=' : tc.TDBQCNUMGE,
        '<' : tc.TDBQCNUMLT,
        '<=' : tc.TDBQCNUMLE,
    },
}    


class BaseField(object):
    INDEX = {}
    CONDITION_TYPE = 'str'
    
    def __init__(self, key=None, default=None, index=False):
        self.key = key
        self.default = default
        self.index = index

    def __get__(self, instance, owner):
        if self.key in instance._data:
            return self.loads(instance._data[self.key])
        return self.default

    def __set__(self, instance, value):
        instance._data[self.key] = self.dumps(value)
        
    def __delete__(self, instance):
        del instance._data[self.key]
        
    def get_condition(self, abbr):
        return CONDITIONS.get(self.CONDITION_TYPE, {})[abbr]
        
    def get_order(self, direction):
        return self.INDEX[direction]
    
    def dumps(self, value):
        return value

    def loads(self, s):
        return s


class ByteStringField(BaseField):
    INDEX = INDEX_STR


class UnicodeField(BaseField):
    '''
        >>> value = 'эээ'.decode('utf-8')
        >>> obj = UnicodeField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    INDEX = INDEX_STR
    
    def __init__(self, internal_encoding='utf-8',
                encoding_errors='ignore', *args, **kwargs):
        self.internal_encoding = internal_encoding
        self.encoding_errors = encoding_errors
        super(UnicodeField, self).__init__(*args, **kwargs)
        
    def dumps(self, value):
        return value.encode(self.internal_encoding, self.encoding_errors)
        
    def loads(self, s):
        return s.decode(self.internal_encoding, self.encoding_errors)


class IntegerField(BaseField):
    '''
        >>> value = 10
        >>> obj = IntegerField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    INDEX = INDEX_NUM
    CONDITION_TYPE = 'num'
    
    def dumps(self, value):
        return str(value)
    
    def loads(self, s):
        return int(s)


class FloatField(BaseField):
    '''
        >>> value = 10.11
        >>> obj = FloatField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    INDEX = INDEX_NUM
    CONDITION_TYPE = 'num'
    
    def dumps(self, value):
        return '%f' % value
    
    def loads(self, s):
        return float(s)


class BooleanField(BaseField):
    '''
        >>> value = True
        >>> obj = BooleanField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    INDEX = INDEX_NUM
    CONDITION_TYPE = 'num'
    
    def dumps(self, value):
        return str(int(value))
    
    def loads(self, s):
        return bool(int(s))


class DateField(BaseField):
    '''
        >>> value = date.today()
        >>> obj = DateField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    INDEX = INDEX_NUM
    CONDITION_TYPE = 'num'
    
    def dumps(self, value):
        return '%i%02i%02i' % (value.year, value.month, value.day)

    def loads(self, s):
        return date(int(s[:-4]), int(s[-4:-2]), int(s[-2:]))


class DateTimeField(BaseField):
    '''
        >>> value = datetime.now()
        >>> obj = DateTimeField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    INDEX = INDEX_NUM
    CONDITION_TYPE = 'num'
    
    def dumps(self, value):
        ms = '' if value.microsecond else '000000'
        return value.isoformat().translate(None, 'T-:.') + ms

    def loads(self, s):
        return datetime(int(s[:-16]), int(s[-16:-14]), int(s[-14:-12]),
            int(s[-12:-10]), int(s[-10:-8]), int(s[-8:-6]), int(s[-6:]))


class DumpField(BaseField):
    '''
        >>> value = [1, {'a': False}]
        >>> obj = DumpField()
        >>> dump = obj.dumps(value)
        >>> isinstance(dump, str)
        True
        >>> obj.loads(dump) == value
        True
    '''
    def __init__(self, dumper='marshal', *args, **kwargs):
        __import__(dumper)
        self.dumper = sys.modules[dumper]
        super(DumpField, self).__init__(*args, **kwargs)
    
    def dumps(self, value):
        return self.dumper.dumps(value)

    def loads(self, s):
        return self.dumper.loads(s)


def _compressed_factory(cls):
    class Ret(cls):
        INDEX = {}
        CONDITION_TYPE = 'str'
        
        def __init__(self, compressor='zlib', compress_level=9, *args, **kwargs):
            __import__(compressor)
            self.compressor = sys.modules[compressor]
            self.compress_level = compress_level
            super(self.__class__, self).__init__(*args, **kwargs)
            
        def dumps(self, value):
            value = self.compressor.compress(value, self.compress_level)
            return super(self.__class__, self).dumps(value)
            
        def loads(self, s):
            s = self.compressor.decompress(s)
            return super(self.__class__, self).loads(value)
            
    Ret.__name__ = 'Compressed' + cls.__name__
    return Ret

CompressedByteStringField = _compressed_factory(ByteStringField)
CompressedUnicodeField = _compressed_factory(UnicodeField)
CompressedDumpField = _compressed_factory(DumpField)


if __name__ == "__main__":
    import doctest
    doctest.testmod()


#!/usr/bin/env python
import os
from setuptools import setup

PACKAGE = 'tclite'
VERSION = '0.1'

readme = open(os.path.join(os.path.dirname(__file__), 'README')).read()

if __name__ == '__main__':
    # Compile the list of packages available, because distutils doesn't have
    # an easy way to do this.
    packages, data_files = [], []
    root_dir = os.path.dirname(__file__)
    if root_dir:
        os.chdir(root_dir)

    for dirpath, dirnames, filenames in os.walk(PACKAGE):
        for i, dirname in enumerate(dirnames):
            if dirname in ['.', '..']:
                del dirnames[i]
        if '__init__.py' in filenames:
            pkg = dirpath.replace(os.path.sep, '.')
            if os.path.altsep:
                pkg = pkg.replace(os.path.altsep, '.')
            packages.append(pkg)
        elif filenames:
            prefix = dirpath[len(PACKAGE) + 1:] # Strip package directory + path separator
            for f in filenames:
                data_files.append(os.path.join(prefix, f))


    setup(
        version = VERSION,
        description = 'Object wrapper for tokyo cabinet',
        long_description = readme,
        author = 'Imbolc',
        author_email = 'imbolc@imbolc.name',
        url = 'http://bitbucket.org/imbolc/tclite/',
        name = PACKAGE,
        #~ requires = ['tokyo-python'],

        packages = packages,
        package_data = {'tclite': data_files},

        license = "BSD",
        keywords = "tc, tokyo, cabinet, orm, model, document, non-relational, document-oriented, key/value",
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )
